local VALUE = 0
local COMBO = 1

local isHorus = (LCD_W == 480)
local pageCount = isHorus and 11 or 7
local edit = false
local page = 1
local current = 1
local refreshState = 0
local pageOffset = 0
local pages = {}
local fields = {}
local modifications = {}
local margin = 1
local spacing = 8

local parameters = {
  -- { name, type, page, sub, value, min, max }
  {"ESC State",                                 COMBO, 0xE0, nil, {"ON", "OFF"}, {1, 0}},

  {"ESC1 Thr CH",                               VALUE, 0xE1, nil, 1, 24, "", 0},
  {"ESC1 PWM Min",                              VALUE, 0xE2, nil, 885, 1500, "Hz", 0},
  {"ESC1 PWM Max",                              VALUE, 0xE3, nil, 1500, 2115, "Hz", 0},

  {"ESC2 Thr CH",                               VALUE, 0xF1, nil, 1, 24, "", 0},
  {"ESC2 PWM Min",                              VALUE, 0xF2, nil, 885, 1500, "Hz", 0},
  {"ESC2 PWM Max",                              VALUE, 0xF3, nil, 1500, 2115, "Hz", 0},

  --{"Soft Brake",          createChoiceField, 0x86, 0, {{"OFF", 0}, {"ON", 1}}},
  --{"3D Mode(Effective after restart)",       createChoiceField, 0x87, 0, {{"OFF", 0}, {"ON", 1}}},
  --{"Current Calibration", createNumberField, 0x88, 100, 0, 100, "%", 0},
  --{"Current Limit",       createNumberField, 0x89, 40, 0, 655, "A"},
  --{"BEC Voltage",         createNumberField, 0x8A, 500, 500, 840, "V", 2},
  --{"Trapezoidal Mode",    createChoiceField, 0x8B, 0, {{"OFF", 0}, {"ON", 1}}},
  --{"Phy Id",              createNumberField, 0x8C, 10, 0, 26},
  --{"App Group Id",        createNumberField, 0x8D, 0, 0, 15},
  --{"Time Gap",            createNumberField, 0x8E, 10, 0, 255, "00ms"},
  --{"Motor Pole Count",    createNumberField, 0x8F, 14, 2, 255},
  --{"FBus Thr CH(Effective after restart)",   createNumberField, 0x90, 1, 1, 255},
  --{"High Demag prot",    createChoiceField, 0x91, 0, {{"OFF", 0}, {"ON", 1}}},
}

-- {"Stabilizing", COMBO, 0x40, 1, nil, {"ON", "OFF"}, {1, 0} },
-- {"AIL Stab Gain", VALUE, 0x44, 2, nil, 0, 200, "%", 0},

-- Change display attribute to current field
local function addField(step)
  local field = fields[current]
  local min, max
  if field[2] == VALUE then
    min = field[5]
    max = field[6]
  elseif field[2] == COMBO then
    min = 1
    max = #(field[5])
  end
  if (step < 0 and field[4] > min) or (step > 0 and field[4] < max) then
    field[4] = field[4] + step
  end
end

-- Select the next or previous page
local function selectPage(step)
  page = 1 + ((page + step - 1 + #pages) % #pages)
  pageOffset = 0
end

-- Select the next or previous editable field
local function selectField(step)
  current = 1 + ((current + step - 1 + #fields) % #fields)
  if current > pageCount + pageOffset then
    pageOffset = current - pageCount
  elseif current <= pageOffset then
    pageOffset = current - 1
  end
end

local function getNextNilField(offset)
  for offsetIndex = offset or 1, #fields do
    if fields[offsetIndex][4] == nil then
      return fields[offsetIndex], offsetIndex
    end
  end
  return nil, nil
end

local function drawProgressBar()
  local finishedCount = 0
  for index, thisField in ipairs(fields) do
    if thisField[4] ~= nil then
      finishedCount = finishedCount + 1
    end
  end
  if LCD_W == 480 then
    local width = (300 * finishedCount) / #fields
    lcd.drawRectangle(106, 10, 300, 6)
    lcd.drawFilledRectangle(108, 12, width, 2);
  else
    local width = (60 * finishedCount) / #fields
    lcd.drawRectangle(51, 1, 60, 6)
    lcd.drawFilledRectangle(53, 3, width, 2);
  end
end

local function drawScreenTitle(title, page, pages)
  lcd.drawFilledRectangle(0, 0, LCD_W, 30, TITLE_BGCOLOR)
  lcd.drawText(1, 5, title, MENU_TITLE_COLOR)
  lcd.drawText(LCD_W - 40, 5, page.."/"..pages, MENU_TITLE_COLOR)
end

-- Redraw the current page
local function redrawFieldsPage()
  lcd.clear()

  if isHorus then
    drawScreenTitle("Rx ESC", page, #pages)
  else
    lcd.drawScreenTitle("Rx ESC", page, #pages)
  end

  if getNextNilField() ~= nil then
    drawProgressBar()
  end

  for index = 1, pageCount, 1 do
    local field = fields[pageOffset + index]
    if field == nil then
      break
    end

    local attr = current == (pageOffset + index) and ((edit == true and BLINK or 0) + INVERS) or 0

    lcd.drawText(1, margin + spacing * index, field[1])

    if field[4] == nil then
      lcd.drawText(LCD_W, margin + spacing * index, "---", attr + RIGHT)
    else
      if field[2] == VALUE then
        lcd.drawText(LCD_W, margin + spacing * index, tostring(field[4]) .. field[7] , attr + RIGHT)
      elseif field[2] == COMBO then
        if field[4] > 0 and field[4] <= #(field[5]) then
          lcd.drawText(LCD_W, margin + spacing * index, field[5][field[4]] , attr + RIGHT)
        end
      end
    end
  end
end

local function telemetryRead(field)
  return sportTelemetryPush(0x17, 0x30, 0x0B5F, field)
end

local function telemetryWrite(field, value)
  return sportTelemetryPush(0x17, 0x31, 0x0B5F, field + value*256)
end

local telemetryPopTimeout = 0
local function refreshNext()
  if refreshState == 0 then
    local thisField = getNextNilField()
    if #modifications > 0 then
      telemetryWrite(modifications[1][1], modifications[1][2])
      modifications[1] = nil
    elseif thisField ~= nil then
      if telemetryRead(thisField[3]) == true then
        refreshState = 1
        telemetryPopTimeout = getTime() + 80 -- normal delay is 500ms
      end
    end
  elseif refreshState == 1 then
    local physicalId, primId, dataId, value = sportTelemetryPop()
    if primId == 0x32 and dataId == 0x0B5F then
      local fieldId = value % 256
      local refreshCount = 0
      -- Check all the fields
      for fieldIndex, thisField in ipairs(fields) do
        if fieldId == thisField[3] then
          refreshCount = refreshCount + 1
          -- Get local value with sub Id
          local fieldValue = math.floor(value / 256)
          -- Set value with checking field type
          if thisField[2] == COMBO and #thisField == 6 then
            for index = 1, #(thisField[6]), 1 do
              if fieldValue == thisField[6][index] then
                thisField[4] = index
                break
              end
            end
          elseif thisField[2] == VALUE and #thisField == 8 then
            thisField[4] = fieldValue - thisField[8]
          end
        end
        if refreshCount >= 3 then
          break
        end
      end
      refreshState = 0
    elseif getTime() > telemetryPopTimeout then
      refreshState = 0
    end
  end
end

local function getFieldValue(field)
  local value = field[4]
  if value == nil then
    return 0
  end
  if field[2] == COMBO and #field == 6 then
    value = field[6][value]
  elseif field[2] == VALUE and #field == 8 then
    value = value + field[8]
  end
  return value
end

local function updateFieldValue()
  local subIdCount = 0
  local value = 0
  for fieldIndex, thisField in ipairs(fields) do
    if fields[current][3] == thisField[3] then
      subIdCount = subIdCount + 1
      local fieldValue = getFieldValue(thisField)
      -- for subId = 2, thisField[4] do
      --   fieldValue = fieldValue * 256
      -- end
      value = value + fieldValue
    end
    if subIdCount >= 3 then
      break
    end
  end
  modifications[#modifications+1] = { fields[current][3], value }
end

-- Main
local function runFieldsPage(event)
  if event == EVT_VIRTUAL_EXIT then -- exit script
    return 2
  elseif event == EVT_VIRTUAL_ENTER then -- toggle editing/selecting current field
    if fields[current][4] ~= nil then
      edit = not edit
      if edit == false then
        updateFieldValue()
      end
    end
  elseif edit then
    if event == EVT_VIRTUAL_INC or event == EVT_VIRTUAL_INC_REPT then
      addField(1)
    elseif event == EVT_VIRTUAL_DEC or event == EVT_VIRTUAL_DEC_REPT then
      addField(-1)
    end
  else
    if event == EVT_VIRTUAL_NEXT then
      selectField(1)
    elseif event == EVT_VIRTUAL_PREV then
      selectField(-1)
    end
  end
  redrawFieldsPage()
  return 0
end

local function runPage(event)
  fields = parameters
  return runFieldsPage(event)
end

-- Init
local function init()
  current, edit, refreshState = 1, false, 0
  modifications = {}
  if LCD_W == 480 then
    margin = 10
    spacing = 20
  end

  pages = {
    runPage,
  }
end

-- Main
local function run(event)
  if event == nil then
    error("Cannot be run as a model script!")
    return 2
  elseif event == EVT_PAGE_BREAK or event == EVT_PAGEDN_FIRST or event == EVT_SHIFT_BREAK then
    selectPage(1)
  elseif event == EVT_PAGE_LONG  or event == EVT_PAGEUP_FIRST or event == EVT_SHIFT_LONG then
    killEvents(event);
    selectPage(-1)
  end

  local result = pages[page](event)
  refreshNext()

  return result
end

return { init=init, run=run }